//##########################################################################################################################
// Filename:
//  lib_usart.c
// Version:     Letzte Aenderung:     Author:
//  API004          17.09.2008          JS
//  MC-TB 0111     ab 25.08.2008       JS
// Modulbeschreibung:
// Usart-ISR's, Treiber, API
// #ifdef __IAR_SYSTEMS_ICC__ fuer GNU-kompatibilitaet eingebaut
// Rx-Timeout-funktionalitaet; z.B. fuer Modbus. App. setzt die Anzahl der ms.
// #ifdef fuer gemeinsamen rx-tx-puffer bei 485-Protokollen eingebaut
// restart ( open) mit Parameteroptionen eingebaut, Int-Status-error korrigiert. 
// Usart 1 (Feldbus Toppy) nur mit TxRdy-Int: Versuch zur Problemloesung der Usart-Blockade.
//##########################################################################################################################
// *************************************************************************************************************************
// Eingebundene Dateien (includes)
// ******************* System-Dateien ********************

#include <string.h>

// ******************* Eigene Dateien ********************
#include "iocan128.h"
#include "lib_USART.h"

// *************************************************************************************************************************
// Definitionen von Konstanten

// *************************************************************************************************************************
// Definitionen von globalen Variablen
volatile unsigned char  UsartRxToCtr[2];   //  Timeoutzaehler in USART-Lib
/* Static Variables */
static unsigned char  UsartIsrRecError[2]; //  Zeichenuebertragungsfehler fuer Applikation sammeln

// UART1
static volatile unsigned char IsrUSART1_RxHead;
static volatile unsigned char USART1_RxTail;
static volatile unsigned char USART1_TxHead;
static volatile unsigned char IsrUSART1_TxTail;
static unsigned char USART1_RxBuf[USART1_RX_BUFFER_SIZE];
#ifdef COMM_RXTX_BUF_1 // gemeinsamer RX-TX-Buffer:
  #define  USART1_TxBuf    USART1_RxBuf   // Tx == Rx
#else // separater Tx Puffer:
  static unsigned char USART1_TxBuf[USART1_TX_BUFFER_SIZE];
#endif


// Test Mm
static volatile unsigned char udr1;
static volatile unsigned char ucsr1a, ucsr1a2;
static volatile unsigned char ucsr1b, ucsr1b2;
static volatile unsigned char ucsr1c, ucsr1c2;
  

#define FOSC           (11059200L) // Hz     CL_FCLK_IO  in canlib:config.h
#define MSK_UART_8BIT           0x06
#define PORT_PIN_LOW      0     // Port-Pin auf Low setzen
#define PORT_PIN_HIGH     1     // Port-Pin auf High setzen
#define PORT_PIN_EINGANG  0     // Port-Pin auf Eingang schalten
#define PORT_PIN_AUSGANG  1     // Port-Pin auf Ausgang schalten
#define PORT_BYTE_EINGANG 0x00  // Port-Byte auf Eingang schalten
#define PORT_BYTE_AUSGANG 0xFF  // Port-Byte auf Ausgang schalten
#define STATUS_OK              0x00  // Kein Fehler aufgetreten
#define ERR_BEREICH            0x11        
#define USART_PAR_NO      0
#define USART_PAR_ODD     1
#define USART_PAR_EVEN    2          // Parity-Konstanten
#define BIT_OUT(port,bit,data)   ((data)?(port|=(1<<bit)):(port&=(~(1<<bit))))
#define DIR485_0_RX   (  BIT_OUT(PORTE,2,PORT_PIN_HIGH), BIT_OUT(DDRE,2,PORT_PIN_AUSGANG) )
#define DIR485_0_TX   (  BIT_OUT(PORTE,2,PORT_PIN_LOW), BIT_OUT(DDRE,2,PORT_PIN_AUSGANG) )
#define DIR485_1_RX   (  BIT_OUT(PORTD,4,PORT_PIN_HIGH), BIT_OUT(DDRD,4,PORT_PIN_AUSGANG) )
#define DIR485_1_TX   (  BIT_OUT(PORTD,4,PORT_PIN_LOW), BIT_OUT(DDRD,4,PORT_PIN_AUSGANG) )
// USART0:
#define UART0_HW_INIT(cfg)   (UCSR0C = cfg)                  // For devices with Extended IO (128)
//#define UART0_HW_INIT(cfg)   (UCSR0C = cfg | (1<<URSEL0) ) // For devices without Extended IO (162)
#define ENABLE_USART_0()     (UCSR0B = ( 1 << RXCIE0 ) | ( 1 << RXEN0 ) | ( 1 << TXEN0 ) )  // ACHTUNG, nicht bei 9-Bit !!
#define DISABLE_USART_0() 	 (UCSR0B &= 0x07)
#define UART0_START_TX()	    (UCSR0B |= ((1<<UDRIE0) | ( 1 << TXEN0 )) )  /* Enable UDRE interrupt (TxBuffer empty)*/
#define UART0_ENABLE_TXRDY()	(UCSR0B |= (1<<TXCIE0))    /* Enable TxReady interrupt */
#define UART0_STOP_TX()	      (UCSR0B &= ~(1<<UDRIE0))   /* Disable UDRE interrupt (TxBuffer empty)*/
#define UART0_DISABLE_RX()	  (UCSR0B &= ~(1<<RXEN0))    /* Disable Rx, flush and clear Rx-Flags*/
#define UART0_DISABLE_TXRDY()	(UCSR0B &= ~(1<<TXCIE0))   /* Disable TxReady interrupt, all send */
#define UART0_RX_EMPTY()   (UCSR0A & (1<<RXC0))

// USART1:
#define SIG_ISR_RX1(x)        // (PORTB_Bit7 ^= 1)
#define SIG_ISR_TX1(x)        // (PORTB_Bit5 ^= 1)
#define SIG_ISR_TX1R(x)       // (PORTB_Bit6 ^= 1)
#define UART1_HW_INIT(cfg)   (UCSR1C = cfg)                  // For devices with Extended IO (128)
//#define UART1_HW_INIT(cfg)   (UCSR1C = cfg | (1<<URSEL1) ) // For devices without Extended IO (162)
#define ENABLE_USART_1()     (UCSR1B = ( 1 << RXCIE1 ) | ( 1 << RXEN1 ) | ( 1 << TXEN1 ) )  // ACHTUNG, nicht bei 9-Bit !!
#define DISABLE_USART_1() 	 (UCSR1B &= 0x07)
#define UART1_START_TX()	    (UCSR1B |= ((1<<UDRIE1) | ( 1 << TXEN1 )) )    /* Enable UDRE interrupt (TxBuffer empty)*/
#define UART1_ENABLE_TXRDY()	(UCSR1B |= (1<<TXCIE1))    /* Enable TxReady interrupt */
#define UART1_STOP_TX()	      (UCSR1B &= ~(1<<UDRIE1))   /* Disable UDRE interrupt (TxBuffer empty)*/
#define UART1_DISABLE_RX()	  (UCSR1B &= ~(1<<RXEN1))    /* Disable Rx, flush and clear Rx-Flags*/
#define UART1_DISABLE_TXRDY()	(UCSR1B &= ~(1<<TXCIE1))   /* Disable TxReady interrupt, all send */
#define UART1_RX_EMPTY()   (UCSR1A & (1<<RXC1))
  
// *************************************************************************************************************************
// Funktionen
/* Prototypes */

///////////////////////////////////////////////////////////////////////////////////////////
//  Funktion:  loescht die Sende und Empfangspuffer der Uart-ComX.
//  Die Uart/Irq muss vorher deaktiviert werden; Sonstige IRQ's oder Taskwechsel sind erlaubt.
//  Wird z.Zt. nur aus RestartAndFlushCom() aufgerufen.
///////////////////////////////////////////////////////////////////////////////////////////
void FlushBuffer_1(void)
{
// ******************** Lokale Variablen ********************
char null = 0;
// ******************** Funktionscode ********************
// 	USART1_RX_DISABLE();    flush Rx-Fifo   Usart ist schon disabled
	USART1_RxTail = null;
	IsrUSART1_RxHead = null;
  memset(USART1_RxBuf,null,sizeof(USART1_RxBuf));
// auch bei gemeinsamen Puffer getrennte Zeiger, ist bersichtlicher
	IsrUSART1_TxTail = null;
	USART1_TxHead = null;
#ifndef COMM_RXTX_BUF_1 // getrennte RX-TX-Buffer: Tx auch flushen
  memset(USART1_TxBuf,null,sizeof(USART1_TxBuf));
#endif
}
/* Initialize USART -Hardware*/
// Int are disabled
void Init_usart0(char refresh)  // struct, vgl. ARM - Lib.
{
  DIR485_0_RX;
  DISABLE_USART_0();
	UsartIsrRecError[0] = 0;    // Fehler Reset
	if(refresh)
    ENABLE_USART_0();           // ACHTUNG  9-Bit !!
	else                        // Kaltstart
  	UART0_HW_INIT(UART0_CFG); // Standard frameformat
}

void Init_usart1(char refresh)
{
  DIR485_1_RX;
  DISABLE_USART_1();
	FlushBuffer_1(); 
	UsartIsrRecError[1] = 0;    // Fehler Reset
	if(refresh)
    ENABLE_USART_1();         // ACHTUNG 9-Bit !!
	else                        // Kaltstart
  	UART1_HW_INIT(UART1_CFG); // Standard frameformat  8data noPar 1Stop
}

unsigned char CloseCom(unsigned char Usart) // disablen
{
unsigned char Status = STATUS_OK;
  if(Usart > 1)
    return(Usart);
//  Disable_interrupt();
  if(Usart == 0)
  {
    DISABLE_USART_0();
    DIR485_0_RX;
  }
  else // if(==1)
  {
    DISABLE_USART_1();
    DIR485_1_RX;
  }
//  Enable_interrupt();
  return(Status);
}

unsigned char RestartCom(unsigned char Usart, unsigned short Baud, unsigned char dataBits, unsigned char Parity, unsigned char Stopbit)
{
  unsigned char Status = STATUS_OK; // ggf. ERR_BEREICH
  unsigned char ucsrnC;             // frame-Register
  if(Usart > 1)
    return(ERR_BEREICH);

//  Disable_interrupt();
// Erweiterung 9 Databit: ENABLE_USART_n(0|1)fuer UCSZn2 in UCSRnB
  ucsrnC = 0;
// if(dataBits>4) (UCSRnC|=(dataBits-5)&0x06)
  ucsrnC |= MSK_UART_8BIT;  // default
  if(Stopbit == 2)
    ucsrnC |= (1<<3); // USBS1 // alles andere = 1Stopp
  if(Parity)
  {
    ucsrnC |= (1<<5); // UPMn1
    if(Parity & USART_PAR_ODD)   //odd(value)?
      ucsrnC |= (1<<4); // UPMn0
  }
// Erweiterung: je nach freq: >19.2bd / 16Mhz: => U2Xn=1 in UCSRnA => Baud/=2;
  Baud = UART_SET_BAUD_WORD(Baud);
  if(Usart == 0)
  {
    Init_usart0(0);        // flush
  	UART0_HW_INIT(ucsrnC); //  frameformat UCSRnC data Par Stop
    UBRR0H = (unsigned char)(Baud >> 8);
    UBRR0L = (unsigned char)(Baud);
    ENABLE_USART_0();      // Achtung 9Bit
  }
  else if(Usart == 1)
  {
//    Init_usart1(0);        // flush
    DIR485_1_RX;
    FlushBuffer_1(); 
    UsartIsrRecError[1] = 0;    // Fehler Reset

  	UART1_HW_INIT(ucsrnC); //  frameformat UCSRnC data Par Stop
    UBRR1H = (unsigned char)(Baud >> 8);
    UBRR1L = (unsigned char)(Baud);
    ENABLE_USART_1();      // Achtung 9Bit
  }
  Enable_interrupt();
  return(Status);
}

unsigned char Usart_GetComError(unsigned char Usart) // Fehler bei der Uebertragung holen
{
  unsigned char tempUint8;
  if(Usart < 2)
  {
//    Disable_interrupt();
    tempUint8 = UsartIsrRecError[Usart];
    UsartIsrRecError[Usart] = 0;       // und loeschen
    Enable_interrupt();
    return(tempUint8);              // direkt zurueck mit Fehlerbits aus UCSRnA
  }
  else
    return(Usart);
}

///////////////////////////////////
/* Interrupt handlers */
////////////// USART0:
#ifdef __IAR_SYSTEMS_ICC__
  #pragma vector=USART0_RX_vect
  __interrupt void USART0_RX_interrupt( void )
#else
  ISR(USART0_RX_vect)
#endif
{
volatile unsigned char tmphead;
}

#ifdef __IAR_SYSTEMS_ICC__
  #pragma vector=USART0_UDRE_vect
  __interrupt void USART0_TX_interrupt( void )
#else
  ISR(USART0_UDRE_vect)
#endif
{
}

#ifdef __IAR_SYSTEMS_ICC__
  #pragma vector=USART0_TX_vect
  __interrupt void USART0_TXrdy_interrupt( void )
#else
  ISR(USART0_TX_vect)
#endif
{
}

///////////////// USART1:

#ifdef __IAR_SYSTEMS_ICC__
  #pragma vector=USART1_RX_vect
  __interrupt void USART1_RX_interrupt( void )
#else
  ISR(USART1_RX_vect)
#endif
{

  // *DEBUG  Test Echo auf COM 0 ausgeben

  ucsr1a2 = UCSR1A;
  ucsr1b2 = UCSR1B;
  ucsr1c2 = UCSR1C;

  udr1 = UDR1;   
  
  if (udr1 == 255)      
  {
    ucsr1a = UCSR1A;
    ucsr1b = UCSR1B;
    ucsr1c = UCSR1C;
    udr1--;      
  }
   
  UDR0 = udr1; // Spiegeln enpfangenes Zeichen auf UART 0
  
  if (udr1 == 4)
    IsrUSART1_RxHead = 0; // Start Zeichen Protokoll  
  else {
    IsrUSART1_RxHead++;
    if (IsrUSART1_RxHead >= USART1_RX_BUFFER_SIZE)
      IsrUSART1_RxHead = 0;
  }
    
  USART1_RxBuf[IsrUSART1_RxHead] = udr1; // Store received data in buffer, clear Status(!)
  
  if (udr1 == 254)      
  {
    ucsr1a++;
    ucsr1b++;
    ucsr1c++;
    ucsr1a2++;
    ucsr1b2++;
    ucsr1c2++;
  }
}

#ifdef __IAR_SYSTEMS_ICC__
  #pragma vector=USART1_UDRE_vect
  __interrupt void USART1_TX_interrupt( void )
#else
  ISR(USART1_UDRE_vect)
#endif
{
	unsigned char tmptail;
SIG_ISR_TX1();
	/* Check if all data is transmitted */
	if ( USART1_TxHead != IsrUSART1_TxTail )
	{
		UDR1 = USART1_TxBuf[IsrUSART1_TxTail];  /* Start transmition */
		// post-increment       Calculate buffer index 
		tmptail = ( IsrUSART1_TxTail + 1 ) & USART1_TX_BUFFER_MASK;
		IsrUSART1_TxTail = tmptail;      /* Store new index */
	}
//	else  // all send  TR1 weg, nur zum Start
	{
		UART1_STOP_TX();          /* Disable UDRE interrupt */
		UART1_ENABLE_TXRDY();     /* Enable Txc interrupt, finish */
	}
}

#ifdef __IAR_SYSTEMS_ICC__
  #pragma vector=USART1_TX_vect
  __interrupt void USART1_TXrdy_interrupt(void)
#else
  ISR(USART1_TX_vect)
#endif
{
	/* Check if all data is transmitted */
	if ( USART1_TxHead != IsrUSART1_TxTail )
	{ 
	  unsigned char tmptail;  // TR1
SIG_ISR_TX1R();
		// aus  USART1_TX_int:  TR1
		UDR1 = USART1_TxBuf[IsrUSART1_TxTail];  /* Start transmition */
		// post-increment       Calculate buffer index 
		tmptail = ( IsrUSART1_TxTail + 1 ) & USART1_TX_BUFFER_MASK;
		IsrUSART1_TxTail = tmptail;      /* Store new index */
	}
	else                             // Tx finished
	{
		DIR485_1_RX;                   // RS485 Direction 'Receive'
SIG_ISR_TX1R();
		UART1_STOP_TX();               // Disable UDRE interrupt (secure)
		UART1_DISABLE_TXRDY();         // Disable Txc  interrupt 
		ENABLE_USART_1();              // wegen evtl. Receive-disable bei gemeinsamen Puffer
	}
}
// end Isr //////////////////////////////////////////////////

//--------------------------------------------------------------------------------------------------------------------------
// Funktionsbeschreibung:
//   Musterfunktion
// bergabe:
//   Var1 -> Erluterung
// Rckgabe:
//   Var1 -> Erluterung
//--------------------------------------------------------------------------------------------------------------------------

///////////////////////////////////////////////////////////////////////////////////////////
// aus MODBUS_User.c file API
// schreibt <nBytes> aus der Pufferadresse <pBuffer>
// an die Kommunikationsschnittstelle <CommnPathNo>
///////////////////////////////////////////////////////////////////////////////////////////
short WriteToCommPath(unsigned char CommnPathNo, unsigned short nBytes, unsigned char * pBuffer) // MBP   X_BYTE
{
volatile unsigned char overflow;
  unsigned char tmphead;
  unsigned char Zaehler;           // Zhlvariable; max. 255 bxtes 0..254 !!

  if(CommnPathNo >1)       // gibts nicht
    return(0);

  if(nBytes > 0)           // mind. 1 Byte
    nBytes = (nBytes ) & 0x00FF; // 8Bit-Anzahl machen, max.255!
  else
    return(0);

  if(CommnPathNo == 1)         // Usart1
  {
    for(Zaehler = 0;          // das erste Element
        Zaehler < nBytes;     // 0..254! bis zum maximalen Index
        Zaehler++)            // alle Elemente durchzhlen
    {
      USART1_TxBuf[USART1_TxHead] = pBuffer[Zaehler]; // in den Sendepuffer das Zeichen schreiben
      tmphead = ( USART1_TxHead + 1 ) & USART1_TX_BUFFER_MASK; // post-increment
      if(tmphead == IsrUSART1_TxTail) // ueberholt
      {
         overflow = 1;        // nicht erhoehen, sonst bricht sendung ab
         nBytes = Zaehler;    // abbrechen, nur soviele werden gesendet
      }
      else
        USART1_TxHead = tmphead;      /* Store new index */
    }
    DIR485_1_TX;                    // RS485-DIRection  = senden;
    UART1_START_TX();               // Tx Interrupt freigeben und TxEnable: INT wenn Tx-Buffer leer
    Enable_interrupt(); 
  }
  return(nBytes);     //  nBytes nur gesendete/kopierte als geschrieben betrachten
}

// Test-Routine
const unsigned char cT1[] = {0x04, 0x30, 0x31, 0x31, 0x39, 0x54, 0x31, 0x30, 0x31, 0x30, 0x32, 0x31, 0x34, 0x30, 0x05, 0x0D, 0x0A};
const unsigned char cT2[] = {0x04, 0x30, 0x31, 0x31, 0x39, 0x54, 0x31, 0x30, 0x31, 0x30, 0x32, 0x31, 0x34, 0x31, 0x05, 0x0D, 0x0A};
const unsigned char cT3[] = {0x04, 0x30, 0x31, 0x31, 0x39, 0x54, 0x31, 0x30, 0x31, 0x30, 0x32, 0x31, 0x34, 0x32, 0x05, 0x0D, 0x0A};

const unsigned char cA1[] = {0x04, 0x30, 0x31, 0x31, 0x39, 0x54, 0x31, 0x30, 0x31, 0x30, 0x32, 0x30, 0x30, 0x31, 0x34, 0x30, 0x02, 0x3D, 0x30, 0x30, 0x30, 0x30, 0x3B, 0x03, 0x30, 0x34, 0x34, 0x35, 0x0D, 0x0A};
const unsigned char cA2[] = {0x04, 0x30, 0x31, 0x31, 0x39, 0x54, 0x31, 0x30, 0x31, 0x30, 0x32, 0x30, 0x30, 0x31, 0x34, 0x31, 0x02, 0x3D, 0x30, 0x30, 0x30, 0x30, 0x3B, 0x03, 0x30, 0x34, 0x34, 0x36, 0x0D, 0x0A};
const unsigned char cA3[] = {0x04, 0x30, 0x31, 0x31, 0x39, 0x54, 0x31, 0x30, 0x31, 0x30, 0x32, 0x30, 0x30, 0x31, 0x34, 0x32, 0x02, 0x3D, 0x30, 0x30, 0x30, 0x30, 0x3B, 0x03, 0x30, 0x34, 0x34, 0x37, 0x0D, 0x0A};

#define L_CT  (sizeof(cT1))
#define L_CA  (sizeof(cA1))

void Test_Mm(void)
{                                  
  if (IsrUSART1_RxHead >= (L_CT - 1))
  {     
    int i;                    
    unsigned char bOK;
    
    if (USART1_RxBuf[IsrUSART1_RxHead] != 0x0A) // LF - Ende Zeichen
      return;
    
    bOK = 1;
    for (i = 0; i < L_CT; i++)
      if (USART1_RxBuf[i] != cT1[i])
      {
        bOK = 0;
        break;
      }     
    if (!bOK)
    {                    
      bOK = 2;
      for (i = 0; i < L_CT; i++)
        if (USART1_RxBuf[i] != cT2[i])
        {
          bOK = 0;
          break;
        }
    }
    if (!bOK)
    {                    
      bOK = 3;
      for (i = 0; i < L_CT; i++)
        if (USART1_RxBuf[i] != cT3[i])
        {
          bOK = 0;
          break;
        }
    }          
    IsrUSART1_RxHead = 0;
    switch (bOK)
    {
      case 1:               
        WriteToCommPath(1, L_CA, (unsigned char *)cA1);
        break;
      case 2:
        WriteToCommPath(1, L_CA, (unsigned char *)cA2);
        break;
      case 3:                    
        WriteToCommPath(1, L_CA, (unsigned char *)cA3);
        break;
    }       
  }
}                                                                                             
